@@ -34,7 +34,10 @@ import traceback
 from types import ModuleType
 import pkg_resources
 
+from typeguard import typechecked
+
 from typing import (
+    cast,
     Any,
     List,
     Optional, Sequence,
@@ -46,10 +49,10 @@ from typing import (
     Union
 )
 
-from enhancements.classproperty import classproperty, ClassPropertyMeta
 from enhancements.exceptions import ModuleFromFileException
 
 
+@typechecked
 def _split_module_string(modulearg: Text, moduleloader: Optional['ModuleParser'] = None) -> Tuple[Text, Text]:
     """split a string in a module/path and the functionname
 
@@ -66,6 +69,7 @@ def _split_module_string(modulearg: Text, moduleloader: Optional['ModuleParser']
     return modname, funcname
 
 
+@typechecked
 def _load_module_from_string(modname: Text, modules_from_file: bool = False) -> ModuleType:
     """Prüfen, ob das Modul von einem Package oder einer Datei geladen werden soll
 
@@ -84,16 +88,17 @@ def _load_module_from_string(modname: Text, modules_from_file: bool = False) ->
         return sys.modules[modname_file]
 
     logging.warning('Loading modules from files is not recommended! Please use a python package instead.')
-    loader = importlib.machinery.SourceFileLoader(modname_file, modname)  # type: ignore
+    loader = importlib.machinery.SourceFileLoader(modname_file, modname)
     module: ModuleType = types.ModuleType(loader.name)
     loader.exec_module(module)
     return module
 
 
+@typechecked
 def _get_valid_module_class(module: ModuleType, funcname: Text) -> Type['BaseModule']:
     """Prüfen, ob das angeforderte Modul existiert und gibt die Klasse zurück
     """
-    handlerclass: Type['BaseModule'] = getattr(module, funcname, None)
+    handlerclass: Type['BaseModule'] = cast(Type['BaseModule'], getattr(module, funcname, None))
     # Prüfen, ob das angeforderte Modul eine Subklasse von BaseModule ist
     if not handlerclass or not isinstance(handlerclass, type) or not issubclass(handlerclass, BaseModule):
         logging.error("BaseModule %s is not subclass of BaseModule!", type(handlerclass))
@@ -101,6 +106,7 @@ def _get_valid_module_class(module: ModuleType, funcname: Text) -> Type['BaseMod
     return handlerclass
 
 
+@typechecked
 def get_module_class(modulelist: Union[Type['BaseModule'], Text, Sequence[Union[Text, Type['BaseModule']]]], moduleloader: Optional['ModuleParser'] = None, modules_from_file: bool = False) -> List[Type['BaseModule']]:
     """Lädt eine Klasse anhand eines Strings.
 
@@ -140,12 +146,16 @@ def get_module_class(modulelist: Union[Type['BaseModule'], Text, Sequence[Union[
         raise ModuleError(message=traceback.format_exc())
     return modules
 
+
+@typechecked
 def load_entry_point(entrypoint: str, name: str) -> Optional[Type['BaseModule']]:
     for entry_point in pkg_resources.iter_entry_points(entrypoint):
         if entry_point.name == name or entry_point.module_name == name:
-            return entry_point.load()
+            return cast(Type['BaseModule'], entry_point.load())
     return None
 
+
+@typechecked
 def load_module(moduleloader: Optional['ModuleParser'] = None, entry_point_name: Optional[str] = None) -> Type['argparse.Action']:
     """Action, um BaseModule mit der Methode "add_module" des ModuleParsers als Kommandozeilenparameter definieren zu können
     """
@@ -176,6 +186,7 @@ def load_module(moduleloader: Optional['ModuleParser'] = None, entry_point_name:
     return ModuleLoaderAction
 
 
+@typechecked
 def append_modules(moduleloader: Optional['ModuleParser'] = None, baseclasses: Optional[Tuple[Type['BaseModule'], ...]] = None, use_entrypoints: bool = False) -> Type['argparse._AppendAction']:
     """Action für den ModuleParser um BaseModule als Kommanozeilen Parameter "--module" definieren zu können
     """
@@ -200,6 +211,7 @@ def append_modules(moduleloader: Optional['ModuleParser'] = None, baseclasses: O
     return ModuleLoaderAppendAction
 
 
+@typechecked
 def get_entrypoint_modules(entry_point_name: Text) -> Dict[Text, Text]:
     entrypoints = {}
     for entry_point in pkg_resources.iter_entry_points(entry_point_name):
@@ -213,7 +225,8 @@ def get_entrypoint_modules(entry_point_name: Text) -> Dict[Text, Text]:
     return entrypoints
 
 
-def set_module_kwargs(entry_point_name, **kwargs: Any) -> Dict[Text, Any]:
+@typechecked
+def set_module_kwargs(entry_point_name: Text, **kwargs: Any) -> Dict[Text, Any]:
     entrypoints = get_entrypoint_modules(entry_point_name)
     if entrypoints:
         kwargs['choices'] = entrypoints.keys()
@@ -224,6 +237,7 @@ def set_module_kwargs(entry_point_name, **kwargs: Any) -> Dict[Text, Any]:
 
 class ModuleError(Exception):
 
+    @typechecked
     def __init__(
         self,
         moduleclass: Optional[Union[Type['BaseModule'], Tuple[Type['BaseModule'], ...]]] = None,
@@ -243,10 +257,12 @@ class InvalidModuleArguments(Exception):
 class _ModuleArgumentParser(argparse.ArgumentParser):
     """Enhanced ArgumentParser to suppress warnings and error during module parsing"""
 
+    @typechecked
     def __init__(self, *args: Any, **kwargs: Any) -> None:
         super().__init__(*args, **kwargs)
         self.exit_on_error = True
 
+    @typechecked
     def error(self, message: Text) -> None:  # type: ignore
         if self.exit_on_error:
             return
@@ -259,6 +275,7 @@ class BaseModule():
     _modules: Optional[List[Tuple[argparse.Action, Any]]] = None
     CONFIG_PREFIX: Optional[Text] = None
 
+    @typechecked
     def __init__(self, args: Optional[Sequence[Text]] = None, namespace: Optional[argparse.Namespace] = None, **kwargs: Any) -> None:
         self.args: argparse.Namespace
         parser_retval = self.parser().parse_known_args(args, namespace)
@@ -277,6 +294,7 @@ class BaseModule():
             setattr(self.args, param_name, param_value)
 
     @classmethod
+    @typechecked
     def add_module(cls, *args: Any, **kwargs: Any) -> None:
         # remove "baseclass" from arguments
         baseclass: Type[BaseModule] = kwargs.pop('baseclass', BaseModule)
@@ -289,25 +307,29 @@ class BaseModule():
             cls.modules().append((cls.parser().add_argument(*args, **set_module_kwargs(baseclass.__name__, **kwargs)), baseclass))
 
     @classmethod
+    @typechecked
     def parser_arguments(cls) -> None:
         pass
 
     @classmethod
+    @typechecked
     def modules(cls) -> List[Tuple[argparse.Action, Any]]:
         if '_modules' not in cls.__dict__ or cls._modules is None:
             cls._modules = []
         return cls._modules
 
     @classmethod
+    @typechecked
     def parser(cls) -> _ModuleArgumentParser:
         if '_parser' not in cls.__dict__:
-            cls._parser = _ModuleArgumentParser(add_help=False, description=cls.__name__)  # type: ignore
+            cls._parser = _ModuleArgumentParser(add_help=False, description=cls.__name__)
             cls.parser_arguments()
         if not cls._parser:
             raise ValueError('failed to create ModuleParser for {}'.format(cls))
         return cls._parser
 
     @classmethod
+    @typechecked
     def argument_group(cls) -> argparse._ArgumentGroup:
         if '_parser_group' not in cls.__dict__:
             parser = cls.parser()
@@ -317,29 +339,11 @@ class BaseModule():
         return cls._parser_group
 
     @classmethod
+    @typechecked
     def config_section_name(cls) -> Text:  # pylint: disable=E0213
         if not cls.CONFIG_PREFIX:
-            return cls.__name__  # type: ignore
-        return "{}:{}".format(cls.CONFIG_PREFIX, cls.__name__)  # type: ignore
-
-
-class LegacyModule(BaseModule, metaclass=ClassPropertyMeta):
-
-    @classproperty
-    def PARSER(cls):
-        return cls.parser()
-
-    @classproperty
-    def MODULES(cls):
-        return cls.modules()
-
-    @classproperty
-    def config_section(cls):
-        return cls.config_section_name()
-
-
-class Module(LegacyModule):
-    pass
+            return cls.__name__
+        return "{}:{}".format(cls.CONFIG_PREFIX, cls.__name__)
 
 
 class ModuleParserPlugin(BaseModule):
@@ -348,13 +352,14 @@ class ModuleParserPlugin(BaseModule):
 
 class ModuleParser(_ModuleArgumentParser):
 
+    @typechecked
     def __init__(
         self,
         default: Optional[Union[Type[BaseModule], Tuple[Type[BaseModule], ...]]] = None,
         baseclass: Optional[Union[Type[BaseModule], Tuple[Type[BaseModule], ...]]] = None,
         baseclass_as_default: bool = True,
         modules_from_file: bool = False,
-        version: Text = None,
+        version: Optional[Text] = None,
         **kwargs: Any
     ) -> None:
         if baseclass is None:
@@ -408,6 +413,7 @@ class ModuleParser(_ModuleArgumentParser):
                 version=self.version
             )
 
+    @typechecked
     def _get_baseclasses(self, baseclass: Union[Type[BaseModule], Tuple[Type[BaseModule], ...]]) -> Tuple[Type[BaseModule], ...]:
         # set default as baseclass if baseclass is not set
         _baseclasses: List[Type[BaseModule]] = list(baseclass) if isinstance(baseclass, tuple) else [baseclass]
@@ -426,11 +432,13 @@ class ModuleParser(_ModuleArgumentParser):
     def parser(self) -> 'ModuleParser':
         return self
 
+    @typechecked
     def add_plugin(self, plugin: Type[ModuleParserPlugin]) -> None:
         if not inspect.isclass(plugin) or not issubclass(plugin, ModuleParserPlugin):
             raise ValueError("plugin must be a class and subclass of BaseModule!")
         self._plugins[plugin] = None
 
+    @typechecked
     def add_parser(self, parser: argparse.ArgumentParser) -> None:
         for module_parser in self._module_parsers:
             if module_parser.description == parser.description:
@@ -440,6 +448,7 @@ class ModuleParser(_ModuleArgumentParser):
         # append parser to list
         self._module_parsers.add(parser)
 
+    @typechecked
     def add_module(self, *args: Any, **kwargs: Any) -> None:
         # remove "baseclass" from arguments
         baseclass = kwargs.pop('baseclass', BaseModule)
@@ -456,26 +465,44 @@ class ModuleParser(_ModuleArgumentParser):
         self._extra_modules.append((self.add_argument(*args, **set_module_kwargs(baseclass.__name__, **kwargs)), baseclass))
         logging.debug("Baseclass: %s", baseclass)
 
+    @typechecked
     def get_module_path(self, module: Text) -> Text:
         return module
 
-    def get_sub_modules(
+    @typechecked
+    def get_sub_modules_args(
         self,
+        *,
         parsed_args: argparse.Namespace,
         args: Optional[Sequence[Text]],
         namespace: Optional[argparse.Namespace],
-        modules: Optional[List[Tuple[argparse.Action, Any]]],
-        use_modules: bool = False
+        modules: List[Tuple[argparse.Action, Type[BaseModule]]],
+    ) -> List[argparse.ArgumentParser]:
+        modulelist = [getattr(parsed_args, m[0].dest) for m in modules if hasattr(parsed_args, m[0].dest)]
+        modulebasecls: List[Tuple[Type[BaseModule], ...]] = [(m[1], ) for m in modules]
+        return self.get_sub_modules(
+            parsed_args=parsed_args,
+            args=args,
+            namespace=namespace,
+            modules=modulelist,
+            baseclasses=modulebasecls
+        )
+
+    @typechecked
+    def get_sub_modules(
+        self,
+        *,
+        parsed_args: argparse.Namespace,
+        args: Optional[Sequence[Text]],
+        namespace: Optional[argparse.Namespace],
+        modules: Optional[List[Type[BaseModule]]],
+        baseclasses: Optional[List[Tuple[Type[BaseModule], ...]]] = None,
     ) -> List[argparse.ArgumentParser]:
         moduleparsers: List[argparse.ArgumentParser] = []
         if not modules:
             return moduleparsers
-        if not use_modules:
-            modulelist = [getattr(parsed_args, m[0].dest) for m in modules if hasattr(parsed_args, m[0].dest)]
-            modulebasecls = [m[1] for m in modules]
-        else:
-            modulelist = [m for m in modules]
-            modulebasecls = [self.baseclasses for _ in modules]
+        modulelist = [m for m in modules]
+        modulebasecls = baseclasses or [self.baseclasses for _ in modules]
 
         for module, baseclass in zip(modulelist, modulebasecls):
             if isinstance(module, str):
@@ -492,15 +519,25 @@ class ModuleParser(_ModuleArgumentParser):
             moduleparsers.append(module.parser())
 
             try:
-                parsed_subargs, _ = module.parser().parse_known_args(args=args, namespace=namespace)
-                moduleparsers.extend(self.get_sub_modules(parsed_subargs, args, namespace, module.modules()))
+                parsed_known_args = module.parser().parse_known_args(args=args, namespace=namespace)
+                if parsed_known_args:
+                    parsed_subargs: argparse.Namespace
+                    parsed_subargs, _ = parsed_known_args
+                    moduleparsers.extend(self.get_sub_modules_args(
+                        parsed_args=parsed_subargs,
+                        args=args,
+                        namespace=namespace,
+                        modules=module.modules()
+                    ))
             except TypeError:
                 logging.exception("Unable to load modules")
         return moduleparsers
 
+    @typechecked
     def _check_value(self, action: Any, value: Any) -> None:
         pass
 
+    @typechecked
     def _create_parser(self, args: Optional[Sequence[Text]] = None, namespace: Optional[argparse.Namespace] = None) -> 'argparse.ArgumentParser':
         parsed_args_tuple = super().parse_known_args(args=args, namespace=namespace)
         if not parsed_args_tuple:
@@ -517,13 +554,25 @@ class ModuleParser(_ModuleArgumentParser):
                 if module is self.baseclasses:
                     raise ModuleError(module, self.baseclasses)
                 self.add_parser(module.parser())
-                parsed_sub_args = module.parser().parse_known_args(args=args, namespace=namespace)
-                submods = self.get_sub_modules(parsed_sub_args, args, namespace, parsed_args.modules, use_modules=True)
-                for submod in submods:
-                    self.add_parser(submod)
+                parsed_sub_args_tuple = module.parser().parse_known_args(args=args, namespace=namespace)
+                if parsed_sub_args_tuple:
+                    parsed_sub_args, _ = parsed_sub_args_tuple
+                    submods = self.get_sub_modules(
+                        parsed_args=parsed_sub_args,
+                        args=args,
+                        namespace=namespace,
+                        modules=parsed_args.modules
+                    )
+                    for submod in submods:
+                        self.add_parser(submod)
 
         # load modules from add_module method
-        moduleparsers = self.get_sub_modules(parsed_args, args, namespace, self._extra_modules)
+        moduleparsers = self.get_sub_modules_args(
+            parsed_args=parsed_args,
+            args=args,
+            namespace=namespace,
+            modules=self._extra_modules
+        )
         for moduleparser in moduleparsers:
             self.add_parser(moduleparser)
 
@@ -541,6 +590,7 @@ class ModuleParser(_ModuleArgumentParser):
         parser = argparse.ArgumentParser(parents=list(self._module_parsers), **self.__kwargs)
         return parser
 
+    @typechecked
     def parse_args(self, args: Optional[Sequence[Text]] = None, namespace: Optional[argparse.Namespace] = None) -> argparse.Namespace:  # type: ignore
         parser = self._create_parser(args=args, namespace=namespace)
         args_namespace = parser.parse_args(args, namespace)
@@ -548,6 +598,7 @@ class ModuleParser(_ModuleArgumentParser):
             return argparse.Namespace()
         return args_namespace
 
+    @typechecked
     def parse_known_args(self, args: Optional[Sequence[Text]] = None, namespace: Optional[argparse.Namespace] = None) -> Tuple[argparse.Namespace, List[str]]:
         parser = self._create_parser(args=args, namespace=namespace)
         return parser.parse_known_args(args, namespace)
